<?php
# This file is part of Website@School, a Content Management System especially designed for schools.
# Copyright (C) 2008-2016 Ingenieursbureau PSD/Peter Fokker <peter@berestijn.nl>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License version 3 as published by
# the Free Software Foundation supplemented with the Additional Terms, as set
# forth in the License Agreement for Website@School (see /program/license.html).
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the License Agreement for Website@School
# along with this program. If not, see http://websiteatschool.eu/license.html

# langdiff.php -- quick and dirty diff between old and current English language files
# Peter Fokker -- 2011-01-12
#
# $Id: langdiff.php,v 1.11 2016/03/23 09:28:20 peter Exp $
#
# See also langdiff.sh
#
# Usage: langdiff.php /path/to/old /path/to/new old_release new_release
#
# 2011-02-14/PF
# We can now do the diff between either a cvs-version or an (official) distribution version.

/** Valid entry points define WASENTRY; prevents direct access to include()'s. */
define('WASENTRY',__FILE__);

$Oldpath = $_SERVER['argv'][1];
$Newpath = $_SERVER['argv'][2];
$Oldrelease = $_SERVER['argv'][3];
$Newrelease = $_SERVER['argv'][4];

$Oldfiles = get_language_files($Oldpath);
$Newfiles = get_language_files($Newpath);

# Read all language files
#
$Old = array();
$New = array();

foreach($Oldfiles as $key => $path) {
    $string = array();
    include $path;
    $Old[$key] = $string;
}
foreach($Newfiles as $key => $path) {
    $string = array();
    include $path;
    $New[$key] = $string;
}

# Open the HTML-page
#
echo <<<EOT
<html>
<head>
<title>Source language changes $Oldrelease - $Newrelease</title>
<style>
a {
  text-decoration: none;
  color: #00F;
}
a:link, a:visited {
  text-decoration: none;
  color: #00F;
}
a:active, a:hover {
  text-decoration: underline;
  color: #F00;
}
</style>
</head>
<body>

<h1>Source language changes $Oldrelease - $Newrelease</h1>

Below is an overview of changes and additions in
release $Newrelease compared to release $Oldrelease.

EOT;

# Now step through all (new) files and (new) strings, collect in $s
#
$total_modified = 0;
$total_new = 0;
$total_unchanged = 0;
$total_total = 0;
$total = array();
$s = '';
foreach($Newfiles as $key => $filename) {
    $i = 0;
    $count_modified = 0;
    $count_new = 0;
    $count_unchanged = 0;
    $s .= sprintf("\n<h2><a name=\"%s\" href=\"#summary\">%s (%s)</a></h2>\n",$key,$New[$key]['translatetool_title'],$key);
    foreach($New[$key] as $k => $v) {
        ++$i;
        if (!isset($Old[$key][$k])) {
            ++$count_new;
            $s .= sprintf("<hr>\n".
                          "<span title=\"%s:%s\"><strong>%d</strong>: *** NEW ***</span><br>\n".
                          "<strong>new</strong>: %s<br>\n",$key,$k,$i,htmlspecialchars($v));
        } elseif ($Old[$key][$k] != $v) {
            ++$count_modified;
            $s .=  sprintf("<hr>\n".
                           "<span title=\"%s:%s\"><strong>%d</strong>: *** MODIFIED ***</span><br>\n".
                           "<strong>old</strong>: %s<br>\n".
                           "<strong>new</strong>: %s<br>\n",
                           $key,$k,$i,htmlspecialchars($Old[$key][$k]),htmlspecialchars($v));
        } else {
            ++$count_unchanged;
        }
    }
    if ($count_modified + $count_new == 0) {
        $s .= "No changes\n";
    } else {
        $s .= sprintf("<hr>\nModified: %d<br>\nNew: %d\n",$count_modified,$count_new);
    }
    $total[$key] = array('new' => $count_new, 
                         'modified' => $count_modified,
                         'unchanged' => $count_unchanged,
                         'total' => $i);
    $total_modified += $count_modified;
    $total_new += $count_new;
    $total_unchanged += $count_unchanged;
    $total_total += $i;
}
# Start with the summary
#
echo "\n<a name=\"summary\" href=\"#summary\"><h2>Summary</h2></a>\n";
if ($total_modified + $total_new == 0) {
    echo "No changes\n";
} else {
    echo "<table border=\"1\" cellpadding=\"5\" style=\"border-collapse: collapse;\">\n".
        " <tr>\n".
        "  <th align=\"left\">File</th>\n".
        "  <th align=\"right\">Modified</th>\n".
        "  <th align=\"right\">New</th>\n".
        "  <th align=\"right\">Unchanged</th>\n".
        "  <th align=\"right\">Total</th>\n".
        " </tr>\n";
    foreach($total as $k => $v) {
        echo " <tr>\n".
            "  <td><a href=\"#$k\">$k</a></td>".
            "  <td align=\"right\">${v['modified']}</td>\n".
            "  <td align=\"right\">${v['new']}</td>\n".
            "  <td align=\"right\">${v['unchanged']}</td>\n".
            "  <td align=\"right\">${v['total']}</td>\n".
            " </tr>\n";
    }
    echo " <tr>\n".
        "  <th align=\"left\">Total</th>".
        "  <th align=\"right\">$total_modified</th>\n".
        "  <th align=\"right\">$total_new</th>\n".
        "  <th align=\"right\">$total_unchanged</th>\n".
        "  <th align=\"right\">$total_total</th>\n".
        " </tr>\n";
    echo "</table>\n";
}
# Now for the details
echo $s;

# Finish HTML-page
#
echo "</body>\n</html>\n";

/** construct a list of files with translations
 *
 * We construct a list of files containing translations, starting in the
 * supplied root directory. We are looking for four or five different sets:
 *
 * ./program/languages/en/*.php  (but not en_manifest.php) (no prefix)
 * ./program/install/languages/en/*.php (prefix 'i_')
 * ./program/themes/*?/languages/en/*php (prefix 't_')
 * ./program/modules/*?/languages/en/*php (prefix 'm_')
 * ./program/addons/*?/languages/en/*php (prefix 'a_')
 *
 * Depending on the supplied root directory we either look at the
 * distribution version (as in the 5 lines in above) OR we have a
 * direct dump from CVS, in which case we are looking for this list:
 *
 * ./languages/en/*.php  (but not en_manifest.php) (no prefix)
 * ./websiteatschool/program/install/languages/en/*.php  (prefix 'i_')
 * ./themes/*?/languages/en/*php (prefix 't_')
 * ./modules/*?/languages/en/*php (prefix 'm_')
 * ./addons/*?/languages/en/*php (prefix 'a_')
 *
 * This routine returns an array with 1 element per file found,
 * keyed by the filename without extension, but with a prefix.
 * The value is the fully qualified path to the file, e.g.:
 *
 * 'was' => '/home/foo/tmp-TwFsXT/new/languages/en/was.php'
 * 'i_demodata' => '/home/foo/tmp-TwFsXT/new/websiteatschool/program/languages/en/demodata.php'
 *
 * @param string $rootdir indicates where to start the search for language files
 * @param string $ll the language to look for (default 'en')
 * @return array with filenames w/o extension and fully qualified paths
 */
function get_language_files($rootdir,$ll='en') {
  $Prog = basename(__FILE__);

  if (is_dir($rootdir.'/program')) { // distribution version
      $wasdirs = array(''=>'/program/languages/'.$ll,'i_'=>'/program/install/languages/'.$ll);
      $tmadirs = array('t_'=>'program/themes','m_'=>'program/modules','a_'=>'/program/addons');
  } else { // cvs-version
      $wasdirs = array(''=>'/languages/'.$ll,'i_'=>'/websiteatschool/program/install/languages/'.$ll);
      $tmadirs = array('t_'=>'/themes','m_'=>'/modules','a_'=>'/addons');
  }
  $files = array();

  foreach($wasdirs as $prefix => $wasdir) {
    $path = $rootdir.$wasdir;
    if (($handle = @opendir($path)) === FALSE) {
      fwrite(STDERR,sprintf("%s: cannot opendir('%s')\n",$Prog,$wasdir));
      continue;
    }
    while (($file = @readdir($handle)) !== FALSE) {
      if (($file == '.') || ($file == '..')) continue;
      if ((substr($file,-4) == '.php') && ($file != $ll.'_manifest.php')) {
	$files[$prefix.substr($file,0,-4)] = sprintf('%s/%s',$path,$file);
      }
    }
    @closedir($handle);
  }
  foreach($tmadirs as $prefix => $tmadir) {
    $path = $rootdir.$tmadir;
    if (($handle = @opendir($path)) === FALSE) {
      fwrite(STDERR,sprintf("%s: cannot opendir('%s')\n",$Prog,$tmadir));
      continue;
    }
    while (($file = @readdir($handle)) !== FALSE) {
      if (($file == '.') || ($file == '..')) continue;
      if ((file_exists(sprintf('%s/%s/languages/%s/%s.php',$path,$file,$ll,$file))) &&
	  (file_exists(sprintf('%s/%s/%s_manifest.php',$path,$file,$file)))) {
	$files[$prefix.$file] = sprintf('%s/%s/languages/%s/%s.php',$path,$file,$ll,$file);
      }
    }
    @closedir($handle);
  }
  return $files;
} // get_language_files()


?>
